import json
import os
from pathlib import Path

CONFIG_FILE_NAME = 'config.json'
ROOT_ENV_NAME = 'JINA_HUB_ROOT'


class Config:
    """
    This class is used to store the configuration of the application.
    """

    def __init__(
        self,
        config_file_name: str = CONFIG_FILE_NAME,
        root_env_name: str = ROOT_ENV_NAME,
    ):
        self.root = Path(os.environ.get(root_env_name, os.path.expanduser('~/.jina')))
        self.config_file = self.root.joinpath(config_file_name)

        if not self.root.exists():
            self.root.mkdir(parents=True, exist_ok=True)

    def get(self, key: str = None):
        """
        Get the value of the key from the config.

        :param key: The key of the config. If it's None, then return the whole config.
        """
        if not self.config_file.exists():
            return None

        with open(self.config_file) as f:
            config = json.load(f)
            if key is None:
                return config
            else:
                return config.get(key)

    def set(self, key: str, value: str):
        """
        Set the value of the key to the config.

        :param key: The key of the config.
        :param value: The value of the config.

        :return: Whole config.
        """
        config = self.get()
        if config is None:
            config = {}

        config[key] = value
        with open(self.config_file, 'w') as f:
            json.dump(config, f)

        return config

    def delete(self, key: str):
        """
        Delete the key from the config.

        :param key: The key of the config.

        :return: Whole config.
        """
        config = self.get()
        if config is None:
            return None

        if key in config:
            del config[key]
            with open(self.config_file, 'w') as f:
                json.dump(config, f)

        return config

    def purge(self):
        """
        Purge the config.
        """
        if self.config_file.exists():
            self.config_file.unlink()


config = Config()



import os
from json import JSONDecodeError

from requests import Response

DOMAIN = 'https://api.hubble.jina.ai'
PROTOCOL = 'rpc'
VERSION = 'v2'


def get_base_url():
    """Get the base url based on environment"""

    domain = DOMAIN
    if 'JINA_HUBBLE_REGISTRY' in os.environ:
        domain = os.environ['JINA_HUBBLE_REGISTRY']

    return f'{domain}/{VERSION}/{PROTOCOL}/'


def get_json_from_response(resp: Response) -> dict:
    """
    Get the JSON data from the response.
    If the response isn't JSON, the response information is lost.
    The error message must include the response body and status code.
    """
    try:
        return resp.json()
    except JSONDecodeError as err:
        raise JSONDecodeError(
            f'Response: {resp.text}, status code: {resp.status_code}; {err.msg}',
            err.doc,
            err.pos,
        ) from err



from hubble.parsers.helper import _chf

from .base import set_base_parser


def mixin_hub_usage_parser(parser):
    """Add the arguments for hub pull to the parser
    :param parser: the parser configure
    """
    parser.add_argument(
        '--no-usage',
        action='store_true',
        default=False,
        help='If set, Hub executor usage will not be printed.',
    )


def set_hub_push_parser(parser=None):
    """Set the parser for the hub push
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from hubble.executor.parsers.push import mixin_hub_push_parser

    mixin_hub_usage_parser(parser)
    mixin_hub_push_parser(parser)
    return parser


def set_hub_pull_parser(parser=None):
    """Set the parser for the hub pull
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from hubble.executor.parsers.pull import mixin_hub_pull_parser

    mixin_hub_usage_parser(parser)
    mixin_hub_pull_parser(parser)
    return parser


def set_hub_new_parser(parser=None):
    """Set the parser for the hub new
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from hubble.executor.parsers.new import mixin_hub_new_parser

    mixin_hub_new_parser(parser)
    return parser


def set_hub_status_parser(parser=None):
    """Set the parser for the hub status
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from hubble.executor.parsers.status import mixin_hub_status_parser

    mixin_hub_status_parser(parser)
    return parser


def set_hub_list_parser(parser=None):
    """Set the parser for the hub list
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from hubble.executor.parsers.list import mixin_hub_list_parser

    mixin_hub_list_parser(parser)
    return parser


def get_main_parser(parser=None):
    """The main parser for Jina Hub CLI
    :return: the parser
    """

    # create the top-level parser
    if not parser:
        parser = set_base_parser()

    sp = parser.add_subparsers(
        dest='hub_cli',
        help='Manage Executor on Jina Hub',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get detailed information about each sub-command',
        required=True,
    )

    set_hub_new_parser(
        sp.add_parser(
            'new',
            help='create a new executor using the template',
            description='Create a new executor using the template',
            formatter_class=_chf,
        )
    )

    set_hub_push_parser(
        sp.add_parser(
            'push',
            help='push an executor package to Jina hub',
            description='Push an executor package to Jina hub',
            formatter_class=_chf,
        )
    )

    set_hub_pull_parser(
        sp.add_parser(
            'pull',
            help='download an executor image/package from Jina hub',
            description='Download an executor image/package from Jina hub',
            formatter_class=_chf,
        )
    )

    set_hub_status_parser(
        sp.add_parser(
            'status',
            help='query an executor building status of of a pushed Executor from Jina hub',
            description='Query an executor building status of of a pushed Executor from Jina hub',
            formatter_class=_chf,
        )
    )

    set_hub_list_parser(
        sp.add_parser(
            'list',
            help='list your local Jina Executors',
            description='List your local Jina Executors',
            formatter_class=_chf,
        )
    )

    return parser



def main():
    from .parsers import get_main_parser

    args = get_main_parser().parse_args()

    try:
        from . import api

        getattr(api, args.auth_cli.replace('-', '_'))(args)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



"""Module wrapping interactions with the local executor packages."""

import json
import os
import shutil
from pathlib import Path
from typing import Dict, Tuple

import yaml
from hubble.executor import HubExecutor
from hubble.executor.helper import (
    __cache_path__,
    get_hub_packages_dir,
    install_requirements,
    is_requirements_installed,
    random_identity,
    unpack_package,
)

SECRET_PATH = 'secrets'


def get_dist_path(uuid: str, tag: str) -> Tuple[Path, Path]:
    """Get the package path according ID and TAG
    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: package and its dist-info path
    """
    pkg_path = get_hub_packages_dir() / uuid
    pkg_dist_path = pkg_path / f'{tag}.dist-info'
    return pkg_path, pkg_dist_path


def get_dist_path_of_executor(executor: 'HubExecutor') -> Tuple[Path, Path]:
    """Return the path of the executor if available.

    :param executor: the executor to check
    :return: the path of the executor package
    """

    pkg_path, pkg_dist_path = get_dist_path(executor.uuid, executor.tag)

    if not pkg_path.exists():
        raise FileNotFoundError(f'{pkg_path} does not exist')
    elif not pkg_dist_path.exists():
        raise FileNotFoundError(f'{pkg_dist_path} does not exist')
    else:
        return pkg_path, pkg_dist_path


def get_config_path(local_id: str) -> 'Path':
    """Get the local configure file
    :param local_id: the random local ID of the executor
    :return: json config path
    """
    return get_hub_packages_dir() / f'{local_id}.json'


def get_lockfile() -> str:
    """Get the path of file locker
    :return: the path of file locker
    """
    return str(get_hub_packages_dir() / 'LOCK')


def get_secret_path(inode: str) -> 'Path':
    """Get the path of secrets
    :param inode: the inode of the executor path
    :return: the path of secrets
    """

    pre_path = Path(f'{__cache_path__}/{SECRET_PATH}/{inode}')

    return pre_path


def load_secret(work_path: 'Path') -> Tuple[str, str]:
    """Get the UUID, Secret and Task id from local cache
    :param work_path: the work path of the executor
    :return: the UUID, secret and task_id
    """

    from cryptography.fernet import Fernet

    preConfig = work_path / '.jina'
    config = get_secret_path(os.stat(work_path).st_ino)
    config.mkdir(parents=True, exist_ok=True)

    local_id_file = config / 'secret.key'
    pre_secret_file = preConfig / 'secret.key'

    if pre_secret_file.exists() and not local_id_file.exists():
        shutil.copyfile(pre_secret_file, local_id_file)

    uuid8 = None
    secret = None
    task_id = None
    if local_id_file.exists():
        with local_id_file.open() as f:
            local_id, local_key = f.readline().strip().split('\t')
            fernet = Fernet(local_key.encode())

        local_config_file = get_config_path(local_id)
        if local_config_file.exists():
            with local_config_file.open() as f:
                local_config = json.load(f)
                uuid8 = local_config.get('uuid8', None)
                encrypted_secret = local_config.get('encrypted_secret', None)
                if encrypted_secret:
                    secret = fernet.decrypt(encrypted_secret.encode()).decode()
                task_id = local_config.get('task_id', None)
    return uuid8, secret, task_id


def dump_secret(work_path: 'Path', uuid8: str, secret: str, task_id: str):
    """Dump the UUID, Secret and Task_id into local file
    :param work_path: the work path of the executor
    :param uuid8: the ID of the executor
    :param secret: the access secret
    :param task_id: the ID of the executor's building task
    """
    from cryptography.fernet import Fernet

    config = get_secret_path(os.stat(work_path).st_ino)
    config.mkdir(parents=True, exist_ok=True)

    local_id_file = config / 'secret.key'
    if local_id_file.exists():
        try:
            with local_id_file.open() as f:
                local_id, local_key = f.readline().strip().split('\t')
                fernet = Fernet(local_key.encode())
        except Exception:
            return
    else:
        local_id = str(random_identity())
        with local_id_file.open('w') as f:
            local_key = Fernet.generate_key()
            fernet = Fernet(local_key)
            f.write(f'{local_id}\t{local_key.decode()}')

    local_config_file = get_config_path(local_id)

    secret_data = {}
    if uuid8:
        secret_data['uuid8'] = uuid8
    if secret:
        secret_data['encrypted_secret'] = fernet.encrypt(secret.encode()).decode()
    if task_id:
        secret_data['task_id'] = task_id

    with local_config_file.open('w') as f:
        f.write(json.dumps(secret_data))


def install_local(
    zip_package: 'Path',
    executor: 'HubExecutor',
    install_deps: bool = False,
):
    """Install the package in zip format to the Jina Hub root.

    :param zip_package: the path of the zip file
    :param executor: the executor to install
    :param install_deps: if set, install dependencies
    """

    pkg_path, pkg_dist_path = get_dist_path(executor.uuid, executor.tag)

    # clean the existed dist_path
    for dist in pkg_path.glob('*.dist-info'):
        shutil.rmtree(dist)

    # unpack the zip package to the root pkg_path
    unpack_package(zip_package, pkg_path)

    # create dist-info folder
    pkg_dist_path.mkdir(parents=False, exist_ok=True)

    install_package_dependencies(install_deps, pkg_dist_path, pkg_path)

    manifest_path = pkg_path / 'manifest.yml'
    if manifest_path.exists():
        shutil.copyfile(manifest_path, pkg_dist_path / 'manifest.yml')

    # store the commit id in local
    if executor.commit_id is not None:
        commit_file = pkg_dist_path / f'PKG-COMMIT-{executor.commit_id}'
        commit_file.touch()


def install_package_dependencies(
    install_deps: bool, pkg_dist_path: 'Path', pkg_path: 'Path'
) -> None:
    """

    :param install_deps: if set, then install dependencies
    :param pkg_dist_path: package distribution path
    :param pkg_path: package path
    """
    # install the dependencies included in requirements.txt
    requirements_file = pkg_path / 'requirements.txt'

    if requirements_file.exists():
        if pkg_path != pkg_dist_path:
            shutil.copyfile(requirements_file, pkg_dist_path / 'requirements.txt')

        if install_deps:
            install_requirements(requirements_file)
        elif not is_requirements_installed(requirements_file, show_warning=True):
            raise ModuleNotFoundError(
                'Dependencies listed in requirements.txt are not all installed locally, '
                'this Executor may not run as expect. To install dependencies, '
                'add `--install-requirements` or set `install_requirements = True`'
            )


def uninstall_local(uuid: str):
    """Uninstall the executor package.

    :param uuid: the UUID of the executor
    """
    pkg_path, _ = get_dist_path(uuid, None)
    for dist in get_hub_packages_dir().glob(f'{uuid}/*.dist-info'):
        shutil.rmtree(dist)
    if pkg_path.exists():
        shutil.rmtree(pkg_path)


def list_local():
    """List the locally-available executor packages.

    :return: the list of local executors (if found)
    """
    result = []
    for dist_name in get_hub_packages_dir().glob(r'*/*.dist-info'):
        result.append(dist_name)

    return result


def exist_local(uuid: str, tag: str = None) -> bool:
    """Check whether the executor exists in local

    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: True if existed, else False
    """
    try:
        get_dist_path_of_executor(HubExecutor(uuid=uuid, tag=tag))
        return True
    except FileNotFoundError:
        return False


def load_config(path: Path) -> Dict:
    """Load config of executor from YAML file.

    :param path: the path of the local executor
    :return: dict
    """
    with open(path / 'config.yml') as fp:
        tmp = yaml.safe_load(fp)

    return tmp


def test_load_config(path: Path) -> Dict:
    """Load config of executor from YAML file.

    :param path: the path of the local executor
    :return: dict
    """
    with open(path / 'config.yml') as fp:
        tmp = yaml.safe_load(fp)

    return tmp



"""Note, this should be consistent with hubble errors."""
from typing import Optional


class BaseError(Exception):
    response = None
    data = {}
    code = -1
    message = "An unknown error occurred"

    def __init__(
        self,
        response: dict,
        data: dict = {},
        message: Optional[str] = None,
        code: Optional[int] = None,
    ):
        self.response = response
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data

    def __str__(self):
        if self.code:
            return f'{self.code}: {self.message}'
        return self.message


class ParamValidationError(BaseError):
    ...


class SQLCreationError(BaseError):
    ...


class DataStreamBrokenError(BaseError):
    ...


class UnexpectedMinetypeError(BaseError):
    ...


class SSOLoginRequiredError(BaseError):
    ...


class AuthenticationFailedError(BaseError):
    ...


class AuthenticationRequiredError(BaseError):
    ...


class OperationNotAllowedError(BaseError):
    ...


class InternalResourceNotFoundError(BaseError):
    ...


class RpcMethodNotFOundError(BaseError):
    ...


class RequestedEntityNotFoundError(BaseError):
    ...


class InternalResourceMethodNotFoundError(BaseError):
    ...


class IncompatiableMethodError(BaseError):
    ...


class InternalResourceIdConflictError(BaseError):
    ...


class FileTooLargeError(BaseError):
    ...


class MaximumUploadSizeReachedError(BaseError):
    ...


class InternalDataCorruptionError(BaseError):
    ...


class IdentifierNamespaceOccupiedError(BaseError):
    ...


class SubmittedDataMalformedError(BaseError):
    ...


class ExternalServiceFailureError(BaseError):
    ...


class ServerInternalError(BaseError):
    ...


class DownstreamServiceError(BaseError):
    ...


class ServerSubprocessError(BaseError):
    ...


class SandboxBuildNotfoundError(BaseError):
    ...


class NotImplementedError(BaseError):
    ...


class ResponseStreamClosedError(BaseError):
    ...


errorcodes = {
    -1: BaseError,
    40001: ParamValidationError,
    40002: SQLCreationError,
    40003: DataStreamBrokenError,
    40004: UnexpectedMinetypeError,
    40101: SSOLoginRequiredError,
    40102: AuthenticationFailedError,
    40103: AuthenticationRequiredError,
    40301: OperationNotAllowedError,
    40401: InternalResourceNotFoundError,
    40402: RpcMethodNotFOundError,
    40403: RequestedEntityNotFoundError,
    40501: InternalResourceMethodNotFoundError,
    40502: IncompatiableMethodError,
    40901: InternalResourceIdConflictError,
    41301: FileTooLargeError,
    41302: MaximumUploadSizeReachedError,
    42201: InternalDataCorruptionError,
    42202: IdentifierNamespaceOccupiedError,
    42203: SubmittedDataMalformedError,
    42204: ExternalServiceFailureError,
    50001: ServerInternalError,
    50002: DownstreamServiceError,
    50003: ServerSubprocessError,
    50004: SandboxBuildNotfoundError,
    50005: NotImplementedError,
    50006: ResponseStreamClosedError,
}



import argparse


def set_base_parser():
    from .helper import _chf

    parser = argparse.ArgumentParser(
        description=f'Jina Auth CLI helps you login in to Jina AI Ecosystem.',  # noqa F501
        formatter_class=_chf,
    )

    return parser



import requests

from ..utils import get_base_url
from .endpoints import EndpointsV2

__all__ = ['HubbleAPISession']


class HubbleAPISession(requests.Session):
    """The customized `requests.Session` object.

    ``HubbleAPISession`` helps the ``hubble.client.Client`` create a default
    ``header`` and validate the jwt token when calling ``init_jwt_auth``.

    The ``HubbleAPISession`` is initialized in the ``hubble.client.Client``
    constructor.
    """

    def __init__(self):
        super().__init__()

        self.headers.update(
            {
                'Accept-Charset': 'utf-8',
            }
        )

    def init_jwt_auth(self, token: str):
        """Initialize the jwt token.

        :param token: The api token user get from webpage.
        """
        self.headers.update({'Authorization': f'token {token}'})

    def validate_token(self) -> requests.Response:
        """Validate API token.

        This function will call the whoami endpoint from Hubble API
        to get user info.

        :return: a `requests.Response` object from Hubble API server.
        """
        url = get_base_url() + EndpointsV2.get_user_info
        resp = requests.post(url, headers=self.headers)
        resp.raise_for_status()

        return resp



import argparse


def set_base_parser():
    from hubble.parsers.helper import _chf

    parser = argparse.ArgumentParser(
        description='Push/Pull an Executor to/from Jina Hub',
        formatter_class=_chf,
    )

    return parser



"""
The Hubble Python Client
"""
import asyncio
import datetime as _datetime
import os as _os
import sys as _sys
from functools import wraps
from typing import Optional

from importlib_metadata import version

from .client.client import Client  # noqa F401
from .excepts import AuthenticationRequiredError
from .utils.auth import Auth  # noqa F401

try:
    __version__ = version("jina-hubble-sdk")
except Exception:
    __version__ = "v0.0.0"


__windows__ = _sys.platform == 'win32'
__uptime__ = _datetime.datetime.now().isoformat()


def login_required(func):
    """Annotate a function so that it requires login to Jina AI to run.

    Example:

    .. highlight:: python
    .. code-block:: python

        @login_required
        def foo():
            print(1)

    :param levels: required build level to run this function.
    :return: annotated function
    """

    @wraps(func)
    def arg_wrapper(*args, **kwargs):
        try:
            Client(jsonify=True).get_user_info()
            return func(*args, **kwargs)
        except AuthenticationRequiredError:
            raise AuthenticationRequiredError(
                response={},
                message=f'Jina auth token is not provided or has expired. {func!r} requires login to Jina AI, '
                f'please do `jina auth login -f` or set env variable `JINA_AUTH_TOKEN` with correct token',
            ) from None

    return arg_wrapper


def login(**kwargs):
    """This function guide user to browser log-in and get token."""
    asyncio.run(Auth.login(**kwargs))


def notebook_login(**kwargs):
    """This function guides user to log-in via P.A.T. or via browser."""
    Auth.login_notebook(**kwargs)


def logout():
    """Logout."""
    asyncio.run(Auth.logout())


def is_logged_in():
    """Check if user is logged in."""
    return True if Client(jsonify=True).token else False


def get_token(interactive: bool = False) -> Optional[str]:
    """Get token."""
    if _os.environ.get('SHOW_HUBBLE_HINT') == 'ALWAYS':
        token = show_hint(interactive)
    elif _os.environ.get('SHOW_HUBBLE_HINT', 'ONCE') == 'ONCE':
        token = show_hint(interactive)
        _os.environ['SHOW_HUBBLE_HINT'] = 'NEVER'
    else:
        token = Client(jsonify=True).token

    return token


def show_hint(interactive: bool = False) -> Optional[str]:  # noqa: E501
    """
    Show hint if the user is not logged in.

    """
    from rich import print

    c = Client(jsonify=True)

    try:
        print(
            f':closed_lock_with_key: [green bold]You are logged in to Jina AI[/] as [bold]{c.username}[/]. '
            f'To log out, use [dim]jina auth logout[/].'
        )
        return c.token
    except AuthenticationRequiredError:
        print(
            ':closed_lock_with_key: [yellow bold]You are not logged in to Jina AI[/]. '
            'To log in, use [bold]jina auth login[/] or set env variable [bold]JINA_AUTH_TOKEN[/].'
        )
        if interactive:
            from rich.prompt import Confirm

            if Confirm.ask('Do you want to login now?'):
                login()

            return show_hint(interactive=interactive)
    except Exception as ex:
        print(f'Unknown error: {ex}')



#!/usr/bin/env python
import json
import os
import sys
from pathlib import Path

from .client.client import Client  # noqa F401


def deploy_hubble_docker_credential_helper_for(*registries: str):
    """
    Deploy hubble docker credential helper for the registry.
    """
    docker_config_dir = os.environ.get('DOCKER_CONFIG', '~/.docker')
    docker_config_dir_path = Path(os.path.expanduser(docker_config_dir))
    docker_config_file_path = Path(
        os.path.expanduser(docker_config_dir + '/config.json')
    )
    target_conf = {}

    if docker_config_file_path.exists():
        with docker_config_file_path.open('r+') as f:
            target_conf = json.load(f)
    if 'credHelpers' not in target_conf:
        target_conf['credHelpers'] = {}
    for registry in registries:
        target_conf['credHelpers'][registry] = 'jina-hubble'

    if not docker_config_dir_path.exists():
        docker_config_dir_path.mkdir(parents=True, exist_ok=True)

    with docker_config_file_path.open('w') as f:
        json.dump(target_conf, f, sort_keys=True, indent=2)
        f.write('\n')


def get_credentials_for(_registry: str):
    """
    Get credentials for the registry.
    """
    token = Client(jsonify=True).token
    username = os.environ.get('HUBBLE_DOCKER_AUTH_OVERRIDE_USERNAME', '<token>')
    secret = os.environ.get('HUBBLE_DOCKER_AUTH_OVERRIDE_SECRET', token)

    sys.stdout.write(
        json.dumps(
            {'Username': username, 'Secret': secret if secret else 'anonymous'},
            indent=2,
        )
    )
    sys.stdout.write('\n')


def auto_deploy_hubble_docker_credential_helper():
    """
    Discover then deploy hubble docker credential helper for internal registries.
    """
    client = Client(jsonify=True)
    registries = client.list_internal_docker_registries().get('data', [])
    deploy_hubble_docker_credential_helper_for(*registries)


def main():
    """
    Main entry point.
    """
    import argparse

    parser = argparse.ArgumentParser(
        description='Hubble docker credential helper for the registry.'
    )
    parser.add_argument(
        'action', nargs='?', default='get', help='Action: get, store, erase, or deploy'
    )
    parser.add_argument(
        'registry', nargs='?', help='The registry to deploy helper for.'
    )

    args = parser.parse_args()

    # sys.stdin.readlines()

    if args.action == 'get':
        get_credentials_for(args.registry)
        sys.exit(0)
    elif args.action == 'deploy':
        if not args.registry:
            auto_deploy_hubble_docker_credential_helper()
            sys.exit(1)
        deploy_hubble_docker_credential_helper_for(args.registry)
        sys.exit(0)
    else:
        sys.exit(1)



import io
import json
from typing import Dict, List, Optional, Union

import requests

from ..utils.api_utils import get_json_from_response
from .base import BaseClient
from .endpoints import EndpointsV2


class Client(BaseClient):
    def create_personal_access_token(
        self, name: str, expiration_days: int = 30
    ) -> Union[requests.Response, dict]:
        """Create a personal access token.

        Personal Access Token (refer as PAT) is same as `token`
        where you get from the UI.
        The main difference is that you can set a ``expiration_days``
        for PAT while ``token`` becomes invalid as soon as user logout.

        :param name: The name of the personal access token.
        :param expiration_days: Number of days to be valid, by default 30 days.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(
            url=self._base_url + EndpointsV2.create_pat,
            data={'name': name, 'expirationDays': expiration_days},
        )

    def list_personal_access_tokens(self) -> Union[requests.Response, dict]:
        """List all created personal access tokens.

        All expired PATs will be automatically deleted.
        The list function only shows valid PATs.

        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(url=self._base_url + EndpointsV2.list_pats)

    def delete_personal_access_token(self, name: str) -> Union[requests.Response, dict]:
        """Delete personal access token by name.

        :param name: Name of the personal access token
          to be deleted.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(
            url=self._base_url + EndpointsV2.delete_pat,
            data={'name': name},
        )

    def get_user_info(self) -> Union[requests.Response, dict]:
        """Get current logged in user information.

        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(url=self._base_url + EndpointsV2.get_user_info)

    @property
    def token(self):
        try:
            self.get_user_info()
            return self._token
        except Exception:
            return None

    @property
    def username(self) -> str:
        return self.get_user_info()['data']['nickname']

    def upload_artifact(
        self,
        f: Union[str, io.BytesIO],
        id: Optional[str] = None,
        metadata: Optional[dict] = None,
        is_public: bool = False,
        show_progress: bool = False,
    ) -> Union[requests.Response, dict]:
        """Upload artifact to Hubble Artifact Storage.

        :param f: The full path or the `io.BytesIO` of the file to be uploaded.
        :param id: Optional value, the id of the artifact.
        :param metadata: Optional value, the metadata of the artifact.
        :param is_public: Optional value, if this artifact is public or not,
          default not public.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        from rich import filesize

        from ..utils.pbar import get_progressbar

        self.get_user_info()  # to make sure the user is logged in.

        pbar = get_progressbar(disable=not show_progress)

        class BufferReader(io.BytesIO):
            def __init__(self, buf=b''):
                super().__init__(buf)
                self._len = len(buf)

                self.task = pbar.add_task(
                    'Uploading',
                    total=self._len,
                    start=True,
                    total_size=str(filesize.decimal(self._len)),
                )

            def __len__(self):
                return self._len

            def read(self, n=-1):
                chunk = io.BytesIO.read(self, n)
                pbar.update(self.task, advance=len(chunk))
                return chunk

        if isinstance(f, str):
            f = open(f, 'rb')
        elif not isinstance(f, io.BytesIO):
            raise TypeError(
                f'Unexpected type {type(f)}, expect either `str` or `io.BytesIO`.'
            )

        dict_data = {
            'public': is_public,
            'file': ('file', f.read()),
        }

        if id:
            dict_data['id'] = id
        if metadata:
            dict_data['metaData'] = json.dumps(metadata)

        (data, ctype) = requests.packages.urllib3.filepost.encode_multipart_formdata(
            dict_data
        )

        headers = {'Content-Type': ctype}

        with pbar:
            return self.handle_request(
                url=self._base_url + EndpointsV2.upload_artifact,
                data=BufferReader(data),
                headers=headers,
            )

    def download_artifact(
        self, id: str, f: Union[str, io.BytesIO], show_progress: bool = False
    ) -> str:
        """Download artifact from Hubble Artifact Storage to localhost.

        :param id: The id of the artifact to be downloaded.
        :param f: The full path or the `io.BytesIO` of the file to be downloaded.
        :returns: A str object indicates the download path on localhost or bytes.
        """
        from rich import filesize

        from ..utils.pbar import get_progressbar

        # first get download uri.
        resp = self.handle_request(
            url=self._base_url + EndpointsV2.download_artifact,
            data={'id': id},
        )

        # Second download artifact.
        if isinstance(resp, requests.Response):
            resp = get_json_from_response(resp)
        download_url = resp['data']['download']

        pbar = get_progressbar(disable=not show_progress)

        with pbar:
            with requests.get(download_url, stream=True) as response:
                total = int(response.headers.get('content-length'))
                task = pbar.add_task(
                    'Downloading',
                    total=total,
                    start=True,
                    total_size=str(filesize.decimal(total)),
                )

                if isinstance(f, str):
                    with open(f, 'wb') as writer:
                        for data in response.iter_content(chunk_size=1024 * 1024):
                            writer.write(data)
                            pbar.update(task, advance=len(data))
                elif isinstance(f, io.BytesIO):
                    for data in response.iter_content(chunk_size=1024 * 1024):
                        f.write(data)
                        pbar.update(task, advance=len(data))
                else:
                    raise TypeError(
                        f'Unexpected type {type(f)}, expect either'
                        '`str` or `io.BytesIO`.'
                    )
        return f

    def delete_artifact(
        self, id: Optional[str] = None, name: Optional[str] = None
    ) -> Union[requests.Response, dict]:
        """Delete the artifact from Hubble Artifact Storage.

        :param id: The id of the artifact to be deleted.
        :param name: The name of the artifact to be deleted.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(
            url=self._base_url + EndpointsV2.delete_artifact,
            data={'id': id} if id else {'name': name},
        )

    def get_artifact_info(
        self, id: Optional[str] = None, name: Optional[str] = None
    ) -> Union[requests.Response, dict]:
        """Get the metadata of the artifact.

        :param id: The id of the artifact to be deleted.
        :param name: The name of the artifact to be deleted.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(
            url=self._base_url + EndpointsV2.get_artifact_info,
            data={'id': id} if id else {'name': name},
        )

    def list_artifacts(
        self,
        *,
        filter: Optional[dict] = None,
        sort: Optional[Dict[str, int]] = None,
        pageIndex: Optional[int] = None,
        pageSize: Optional[int] = None,
    ) -> Union[requests.Response, List[dict]]:
        """Get list of artifacts.

        :param filter: optional, to filter by fields.
        :param sort: optional, to sort by fields.
        :param pageIndex: optional, to specify which page to load.
        :param pageSize: optional, number of items per page.
        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """

        data = {
            'filter': filter,
            'sort': sort,
            'pageIndex': pageIndex,
            'pageSize': pageSize,
        }

        return self.handle_request(
            url=self._base_url + EndpointsV2.list_artifacts,
            json={key: value for (key, value) in data.items() if value is not None},
        )

    def list_internal_docker_registries(
        self,
    ) -> Union[requests.Response, dict]:
        """List internal docker registries.

        :returns: `requests.Response` object as returned value
            or indented json if jsonify.
        """
        return self.handle_request(
            url=self._base_url + EndpointsV2.list_internal_docker_registries,
        )



"""Module for wrapping Jina Hub API calls."""

import argparse
import copy
import hashlib
import json
import logging
import os
import random
from pathlib import Path
from typing import Dict, List, Optional, Union
from urllib.parse import urljoin

import hubble
from hubble.executor import HubExecutor
from hubble.executor.helper import (
    __resources_path__,
    __unset_msg__,
    archive_package,
    check_requirements_env_variable,
    disk_cache_offline,
    download_with_resume,
    get_cache_db,
    get_download_cache_dir,
    get_hub_packages_dir,
    get_hubble_error_message,
    get_request_header,
    get_requirements_env_variables,
    get_rich_console,
    get_tag_from_dist_info_path,
    parse_hub_uri,
    retry,
    status_task,
    upload_file,
)
from hubble.executor.hubapi import (
    dump_secret,
    get_dist_path_of_executor,
    get_lockfile,
    install_local,
    install_package_dependencies,
    list_local,
    load_config,
    load_secret,
)
from hubble.utils.api_utils import get_json_from_response


class HubIO:
    """:class:`HubIO` lets you interact with the Jina Hub registry.
    You can use it from the CLI to package a directory into a Jina Hub Executor and publish it to the world.
    Examples:
        - :command:`jina hub push my_executor/` to push the Executor package to Jina Hub
        - :command:`jina hub pull <UUID8>` to download the Executor identified by its UUID

    To create a :class:`HubIO` object, simply:

        .. highlight:: python
        .. code-block:: python
            hubio = HubIO(args)

    :param args: arguments
    """

    def __init__(self, args: argparse.Namespace):
        self.args = args
        self.logger = logging.getLogger(self.__class__.__name__)

    def new(self) -> None:
        """Create a new Executor folder interactively."""

        from rich import box, print
        from rich.panel import Panel
        from rich.progress import track
        from rich.prompt import Confirm, Prompt
        from rich.syntax import Syntax
        from rich.table import Table

        console = get_rich_console()

        print(
            Panel.fit(
                '''
An [bold green]Executor[/bold green] is how Jina processes [bold]Documents[/bold].

This guide helps you create an Executor in 30 seconds.''',
                title='Create New Executor',
            )
        )

        exec_name = (
            self.args.name
            if self.args.name
            else Prompt.ask(
                ':grey_question: What is the [bold]name[/bold] of your Executor?\n'
                '[dim]CamelCase is required[/dim]',
                default=f'MyExecutor{random.randint(0, 100)}',
            )
        )

        exec_path = (
            self.args.path
            if self.args.path
            else Prompt.ask(
                ':grey_question: Which [bold]folder[/bold] do you want to store your Executor in?',
                default=os.path.join(os.getcwd(), exec_name),
            )
        )
        exec_description = '{{}}'
        exec_keywords = '{{}}'
        exec_url = '{{}}'

        is_dockerfile = 'none'

        if self.args.advance_configuration or Confirm.ask(
            '[green]That\'s all we need to create an Executor![/green]\n'
            ':grey_question: Or do you want to proceed to advanced configuration? '
            '[dim](GPU support, meta information on Hub, etc.)[/]',
            default=False,
        ):
            print(
                Panel.fit(
                    '''
[bold]Dockerfile[/bold] describes how this Executor will be built. It is useful when
your Executor has non-trivial dependencies or must be run in a certain environment.

- If [bold]Dockerfile[/bold] is not given, Jina Cloud automatically generates one.
- If you provide a [bold]Dockerfile[/bold], then Jina Cloud will respect it when building the Executor.

Here are some Dockerfile templates to choose from:
- [b]cpu[/b]: CPU-only Executor with Jina as base image;
- [b]torch-gpu[/b]: GPU-enabled Executor with PyTorch as the base image;
- [b]tf-gpu[/b]: GPU-enabled Executor with TensorFlow as the base image;
- [b]jax-gpu[/b]: GPU-enabled Executor with JAX installed.
''',
                    title=':package: [bold]Dockerfile[/bold]',
                    width=80,
                )
            )

            is_dockerfile = self.args.dockerfile or Prompt.ask(
                ':grey_question: How do you want to generate the [bold]Dockerfile[/bold] for this Executor?',
                choices=['cpu', 'torch-gpu', 'tf-gpu', 'jax-gpu', 'none'],
                default='cpu',
            )

            print(
                Panel.fit(
                    '''
Metadata helps other users identify, search and reuse your Executor on Jina Cloud.
''',
                    title=':name_badge: [bold]Meta Info[/bold]',
                    width=80,
                )
            )

            exec_description = (
                self.args.description
                if self.args.description
                else (
                    Prompt.ask(
                        ':grey_question: Please give a [bold]short description[/bold] of your Executor:\n'
                        f'[dim]Example: {exec_name} embeds images into 128-dim vectors using ResNet.[/dim]'
                    )
                )
            )

            exec_keywords = (
                self.args.keywords
                if self.args.keywords
                else (
                    Prompt.ask(
                        ':grey_question: Please give some [bold]keywords[/bold] '
                        'to help people search your Executor [dim](separated by commas)[/dim]\n'
                        '[dim]Example: image, cv, embedding, encoding, resnet[/dim]'
                    )
                )
            )

            exec_url = (
                self.args.url
                if self.args.url
                else (
                    Prompt.ask(
                        ':grey_question: What is the [bold]URL[/bold] of the Executor\'s GitHub repo?\n'
                        '[dim]Example: https://github.com/yourname/my-executor[/dim]'
                    )
                )
            )

            print('[green]That\'s all we need to create an Executor![/green]')

        def mustache_repl(srcs):
            for src in track(
                srcs, description=f'Creating {exec_name}...', total=len(srcs)
            ):
                dest = src
                if dest.endswith('.Dockerfile'):
                    dest = 'Dockerfile'
                with open(
                    os.path.join(__resources_path__, 'executor-template', src)
                ) as fp, open(os.path.join(exec_path, dest), 'w') as fpw:
                    f = (
                        fp.read()
                        .replace('{{exec_name}}', exec_name)
                        .replace(
                            '{{exec_description}}',
                            exec_description if exec_description != '{{}}' else '',
                        )
                        .replace(
                            '{{exec_keywords}}',
                            str(exec_keywords.split(','))
                            if exec_keywords != '{{}}'
                            else '[]',
                        )
                        .replace('{{exec_url}}', exec_url if exec_url != '{{}}' else '')
                    )
                    fpw.writelines(f)

        Path(exec_path).mkdir(parents=True, exist_ok=True)
        pkg_files = [
            'executor.py',
            'README.md',
            'requirements.txt',
            'config.yml',
        ]

        if is_dockerfile == 'cpu':
            pkg_files.append('Dockerfile')
        elif is_dockerfile == 'torch-gpu':
            pkg_files.append('torch.Dockerfile')
        elif is_dockerfile == 'jax-gpu':
            pkg_files.append('torch.Dockerfile')
        elif is_dockerfile == 'tf-gpu':
            pkg_files.append('tf.Dockerfile')
        elif is_dockerfile != 'none':
            raise ValueError(f'Unknown Dockerfile type: {is_dockerfile}')

        mustache_repl(pkg_files)

        table = Table(box=box.SIMPLE)
        table.add_column('Filename', style='cyan', no_wrap=True)
        table.add_column('Description', no_wrap=True)

        # adding the columns in order of `ls` output
        table.add_row(
            'config.yml',
            'The YAML configuration file of the Executor. You can define [bold]__init__[/bold] '
            'arguments using the [bold]with[/bold] keyword.'
            + '\nYou can also define metadata for the Executor, for easier discovery on Jina Hub.',
        )

        table.add_row(
            '',
            Panel(
                Syntax(
                    f'''
jtype: {exec_name}
with:
  foo: 1
  bar: hello
py_modules:
  - executor.py
metas:
  name: {exec_name}
  description: {exec_description if exec_description != '{{}}' else ''}
  url: {exec_url if exec_url != '{{}}' else ''}
  keywords: {exec_keywords if exec_keywords != '{{}}' else '[]'}
''',
                    'yaml',
                    theme='monokai',
                    line_numbers=True,
                    word_wrap=True,
                ),
                title='config.yml',
                width=50,
                expand=False,
            ),
        )

        if is_dockerfile != 'none':
            table.add_row(
                'Dockerfile',
                'The Dockerfile describes how this Executor will be built.',
            )

        table.add_row('executor.py', 'The Executor\'s main logic file.')
        table.add_row('README.md', 'The Executor\'s usage guide.')
        table.add_row('requirements.txt', 'The Executor\'s Python dependencies.')

        final_table = Table(box=None)

        final_table.add_row(
            'Congratulations! You have successfully created an Executor! Here are the next steps:'
        )

        p0 = Panel(
            Syntax(
                f'ls {exec_path}',
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='1. Check out your Executor',
            width=120,
            expand=False,
        )

        p1 = Panel(
            table,
            title='2. Understand Executor folder structure',
            width=120,
            expand=False,
        )

        p12 = Panel(
            Syntax(
                f'jina executor --uses {exec_path}/config.yml',
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='3. Test your Executor locally',
            width=120,
            expand=False,
        )

        p2 = Panel(
            Syntax(
                f'jina hub push {exec_path}',
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='4. Share your Executor on Jina Hub',
            width=120,
            expand=False,
        )

        for _p in [p0, p1, p12, p2]:
            final_table.add_row(_p)

        p = Panel(
            final_table,
            title=':tada: Next steps',
            width=130,
            expand=False,
        )
        console.print(p)

    def _async_push(
        self,
        console,
        st,
        req_header,
        content,
        form_data,
        work_path,
        task_id,
    ):

        verbose = form_data.get('verbose', False)
        session_id = req_header.get('jinameta-session-id')

        if task_id and not form_data.get('id'):
            image = self._status_with_progress(console, st, task_id, False, False)

            if image:
                new_uuid8 = image['id']
                form_data['id'] = new_uuid8

                dump_secret(work_path, new_uuid8, form_data.get('secret'), task_id)
            else:
                raise Exception(f'Unknown Error, session_id: {session_id}')

        if form_data.get('id'):
            hubble_url = urljoin(hubble.utils.get_base_url(), 'executor.updateAsync')
        else:
            hubble_url = urljoin(hubble.utils.get_base_url(), 'executor.createAsync')

        # upload the archived Executor to Jina Hub
        st.update('Async Uploading...')
        resp = upload_file(
            hubble_url,
            'filename',
            content,
            dict_data=form_data,
            headers=req_header,
            stream=True,
            method='post',
        )

        verbose = form_data.get('verbose', False)
        push_task = None
        for stream_line in resp.iter_lines():
            stream_msg = json.loads(stream_line)
            if stream_msg.get('code') == 202:
                push_task = stream_msg['data']
            else:
                msg = stream_msg.get('message')
                raise Exception(f'{ msg or "Unknown Error"} session_id: {session_id}')

        new_task_id = push_task.get('_id')
        if new_task_id:
            dump_secret(
                work_path, form_data.get('id'), form_data.get('secret'), new_task_id
            )
            self._prettyprint_status_usage(console, work_path, new_task_id)
            st.update('Async Uploaded!')

            image = self._status_with_progress(console, st, new_task_id, False, verbose)
            if image:
                # `new_secret` is always None
                new_uuid8, new_secret = self._prettyprint_result(console, image)
                if new_uuid8 != form_data.get('id'):
                    dump_secret(
                        work_path, new_uuid8, form_data.get('secret'), new_task_id
                    )
            else:
                raise Exception(f'Unknown Error, session_id: {session_id}')

        else:
            raise Exception(f'Error: can\'t get task_id session_id: {session_id}')

    def _sync_push(
        self, console, st, req_header, content, form_data, work_path, uuid8, secret
    ):
        # upload the archived Executor to Jina Hub
        if form_data.get('id'):
            hubble_url = urljoin(hubble.utils.get_base_url(), 'executor.update')
        else:
            hubble_url = urljoin(hubble.utils.get_base_url(), 'executor.create')

        st.update('Uploading...')
        resp = upload_file(
            hubble_url,
            'filename',
            content,
            dict_data=form_data,
            headers=req_header,
            stream=True,
            method='post',
        )

        warnings = []
        verbose = form_data.get('verbose', False)
        image = None
        session_id = req_header.get('jinameta-session-id')
        for stream_line in resp.iter_lines():

            stream_msg = json.loads(stream_line)
            t = stream_msg.get('type')
            subject = stream_msg.get('subject')
            payload = stream_msg.get('payload', '')

            if t == 'error':
                msg = stream_msg.get('message')
                hubble_err = payload
                overridden_msg = ''
                detail_msg = ''
                if isinstance(hubble_err, dict):
                    (overridden_msg, detail_msg) = get_hubble_error_message(hubble_err)
                    if not msg:
                        msg = detail_msg

                if overridden_msg and overridden_msg != detail_msg:
                    self.logger.warning(overridden_msg)

                raise Exception(
                    f'{overridden_msg or msg or "Unknown Error"} session_id: {session_id}'
                )
            elif t == 'warning':
                warnings.append(stream_msg.get('message'))

            if t == 'progress' and subject == 'buildWorkspace':
                legacy_message = stream_msg.get('legacyMessage', {})
                status = legacy_message.get('status', '')
                st.update(f'Cloud building ... [dim]{subject}: {t} ({status})[/dim]')

            elif t == 'complete':
                image = stream_msg['payload']
                warnings.append(stream_msg.get('warning'))
                st.update(
                    f'Cloud building ... [dim]{subject}: {t} ({stream_msg["message"]})[/dim]'
                )
                break

            elif t and subject:
                if verbose and t == 'console':
                    console.log(f'Cloud building ... [dim]{subject}: {payload}[/dim]')
                else:
                    st.update(f'Cloud building ... [dim]{subject}: {t} {payload}[/dim]')

        if image:
            new_uuid8, new_secret = self._prettyprint_result(
                console, image, warnings=warnings
            )
            if new_uuid8 != uuid8 or new_secret != secret:
                dump_secret(work_path, new_uuid8, new_secret or '', None)
        else:
            raise Exception(f'Unknown Error, session_id: {session_id}')

    def push(self) -> None:
        """Push the Executor package to Jina Hub."""

        work_path = Path(self.args.path)

        exec_tags = None
        exec_immutable_tags = None
        if self.args.tag:
            exec_tags = ','.join(self.args.tag)
        if self.args.protected_tag:
            exec_immutable_tags = ','.join(self.args.protected_tag)

        dockerfile = None
        if self.args.dockerfile:
            dockerfile = Path(self.args.dockerfile)
            if not dockerfile.exists():
                raise Exception(f'The given Dockerfile `{dockerfile}` does not exist!')
            if dockerfile.parent != work_path:
                raise Exception(
                    f'The Dockerfile must be placed at the given folder `{work_path}`'
                )

            dockerfile = dockerfile.relative_to(work_path)

        build_env = None
        if self.args.build_env:
            build_envs = self.args.build_env.strip().split()
            build_env_dict = {}
            for index, env in enumerate(build_envs):
                env_list = env.split('=')
                if len(env_list) != 2:
                    raise Exception(
                        f'The `--build-env` parameter: `{env}` is in the wrong format. '
                        f'you can use: `--build-env {env}=YOUR_VALUE`.'
                    )
                if check_requirements_env_variable(env_list[0]) is False:
                    raise Exception(
                        f'The `--build-env` parameter key:`{env_list[0]}` can only '
                        'consist of numbers, upper-case letters and underscore.'
                    )
                build_env_dict[env_list[0]] = env_list[1]
            build_env = build_env_dict if build_env_dict else None

        requirements_file = work_path / 'requirements.txt'

        requirements_env_variables = []
        if requirements_file.exists():
            requirements_env_variables = get_requirements_env_variables(
                requirements_file
            )
            for index, env in enumerate(requirements_env_variables):
                if check_requirements_env_variable(env) is False:
                    raise Exception(
                        f'The requirements.txt environment variables:`${env}` '
                        'can only consist of numbers, upper-case letter and underscore.'
                    )

        if len(requirements_env_variables) and not build_env:
            env_variables_str = ','.join(requirements_env_variables)
            error_str = (
                'requirements.txt sets environment variables as follows:'
                f'`{env_variables_str}` should use `--build-env'
            )
            for item in requirements_env_variables:
                error_str += f' {item}=YOUR_VALUE'
            raise Exception(f'{error_str}` to add it.')
        elif len(requirements_env_variables) and build_env:
            build_env_keys = list(build_env.keys())
            diff_env_variables = list(
                set(requirements_env_variables).difference(set(build_env_keys))
            )
            if len(diff_env_variables):
                diff_env_variables_str = ",".join(diff_env_variables)
                error_str = (
                    'requirements.txt sets environment variables as follows:'
                    f'`{diff_env_variables_str}` should use `--build-env'
                )
                for item in diff_env_variables:
                    error_str += f' {item}=YOUR_VALUE'
                raise Exception(f'{error_str}` to add it.')

        console = get_rich_console()
        with console.status(f'Pushing `{self.args.path}` ...') as st:
            req_header = get_request_header()
            try:
                st.update(f'Packaging {self.args.path} ...')
                md5_hash = hashlib.md5()
                bytesio = archive_package(work_path)
                content = bytesio.getvalue()
                md5_hash.update(content)
                md5_digest = md5_hash.hexdigest()

                # upload the archived package
                form_data = {
                    'public': 'True' if getattr(self.args, 'public', None) else 'False',
                    'private': 'True'
                    if getattr(self.args, 'private', None)
                    else 'False',
                    'md5sum': md5_digest,
                }

                if self.args.verbose:
                    form_data['verbose'] = 'True'

                if self.args.no_cache:
                    form_data['buildWithNoCache'] = 'True'

                if exec_tags:
                    form_data['tags'] = exec_tags

                if exec_immutable_tags:
                    form_data['immutableTags'] = exec_immutable_tags

                if dockerfile:
                    form_data['dockerfile'] = str(dockerfile)

                if build_env:
                    form_data['buildEnv'] = json.dumps(build_env)

                uuid8, secret, task_id = load_secret(work_path)
                if self.args.force_update or uuid8:
                    form_data['id'] = self.args.force_update or uuid8
                if self.args.secret or secret:
                    form_data['secret'] = self.args.secret or secret

                st.update('Connecting to Jina Hub ...')

                logged_in = True if hubble.is_logged_in() else False
                if logged_in:  # if user logged in use async upload api
                    self._async_push(
                        console,
                        st,
                        req_header,
                        content,
                        form_data,
                        work_path,
                        task_id,
                    )

                else:
                    # if user not logged in use async upload api
                    self._sync_push(
                        console,
                        st,
                        req_header,
                        content,
                        form_data,
                        work_path,
                        uuid8,
                        secret,
                    )

            except KeyboardInterrupt:
                pass

            except Exception as e:  # IO related errors
                self.logger.error(
                    f'Please report this session_id: [yellow bold]{req_header["jinameta-session-id"]}[/] '
                    'to https://github.com/floraltin/floralatin-hubble-sdk5/issues'
                )
                raise e

    def _prettyprint_result(
        self, console, image, *, warnings: Optional[List[str]] = None
    ):
        from rich import box
        from rich.panel import Panel
        from rich.table import Table

        uuid8 = image['id']
        secret = image.get('secret')
        visibility = image['visibility']
        commit = image.get('commit', {})
        tags = commit.get('tags', None)
        tag = tags[0] if tags and isinstance(tags, list) else None

        table = Table(box=box.SIMPLE, show_header=False)
        table.add_column(no_wrap=True)
        table.add_column(no_wrap=True)
        if 'name' in image:
            table.add_row(':name_badge: Name', image['name'])

        table.add_row(
            ':link: Jina Hub URL',
            f'[link=https://hub.jina.ai/executor/{uuid8}/]https://hub.jina.ai/executor/{uuid8}/[/link]',
        )

        if secret:
            table.add_row(':lock: Secret', secret)
            table.add_row(
                '',
                '👆 [bold red]Please keep this token in a safe place!',
            )

        table.add_row(':eyes: Visibility', visibility)

        if warnings:
            table.add_row(
                ':exclamation: Warnings',
                '👇 [bold yellow]Process finished with warnings!',
            )
            for warning in warnings:
                table.add_row('', f'[yellow]• {warning}')

        p1 = Panel(
            table,
            title='Published',
            width=100,
            expand=False,
        )
        console.print(p1)

        presented_id = image.get('name', uuid8)
        usage = (
            presented_id
            if visibility == 'public' or not secret
            else f'{presented_id}:{secret}'
        ) + (f'/{tag}' if tag else '')

        if not self.args.no_usage:
            self._prettyprint_usage(console, usage)

        return uuid8, secret

    def _prettyprint_usage(self, console, executor_name):
        from rich import box
        from rich.panel import Panel
        from rich.syntax import Syntax
        from rich.table import Table

        param_str = Table(
            box=box.SIMPLE,
            show_header=False,
        )
        param_str.add_column('')
        param_str.add_column('')
        param_str.add_column('')

        param_str.add_row(
            'Container',
            'YAML',
            Syntax(f"uses: jinahub+docker://{executor_name}", 'yaml'),
        )
        param_str.add_row(
            None,
            'Python',
            Syntax(f".add(uses='jinahub+docker://{executor_name}')", 'python'),
        )

        param_str.add_row()
        param_str.add_row(
            'Sandbox',
            'YAML',
            Syntax(f"uses: jinahub+sandbox://{executor_name}", 'yaml'),
        )
        param_str.add_row(
            '',
            'Python',
            Syntax(f".add(uses='jinahub+sandbox://{executor_name}')", 'python'),
        )

        param_str.add_row()
        param_str.add_row(
            'Source',
            'YAML',
            Syntax(f"uses: jinahub://{executor_name}", 'yaml'),
        )
        param_str.add_row(
            '',
            'Python',
            Syntax(f".add(uses='jinahub://{executor_name}')", 'python'),
        )

        console.print(
            Panel(param_str, title='Usage', expand=False, width=len(executor_name) + 65)
        )

    def _prettyprint_build_env_usage(self, console, build_env, usage_kind=None):
        from rich import box
        from rich.panel import Panel
        from rich.table import Table

        param_str = Table(
            box=box.SIMPLE,
        )
        param_str.add_column('Environment variable')
        param_str.add_column('Your value')

        for index, item in enumerate(build_env):
            param_str.add_row(f'{item}', 'your value')

        console.print(
            Panel(
                param_str,
                title='build_env',
                subtitle='You have to set the above environment variables',
                expand=False,
                width=100,
            )
        )

    def _prettyprint_status_usage(
        self, console, work_path, task_id=None, usage_kind=None
    ):
        from rich import box
        from rich.panel import Panel
        from rich.table import Table

        param_str = Table(
            box=box.SIMPLE,
        )

        param_str.add_column(
            'If you don\'t want to wait for the Executor\'s build status, you can use \'Ctrl + C\'. '
            'After that you can get the status of the Executor any time using the command.'
        )
        param_str.add_row(
            f'Use `jina hub status {work_path}` to get the last build status of the Executor',
        )

        if task_id:
            param_str.add_row('')
            param_str.add_row(
                f'You can also use `jina hub status --id {task_id}` '
                'to get the specified build status of the Executor.',
            )

        console.print(
            Panel(
                param_str,
                title='Status',
                expand=False,
                width=100,
            )
        )

    def _status_with_progress(self, console, st, task_id, replay=False, verbose=False):

        req_header = get_request_header()
        dict_data = {}
        dict_data['replay'] = replay
        dict_data['verbose'] = verbose

        session_id = req_header.get('jinameta-session-id')
        task_progress = status_task(
            url=urljoin(hubble.utils.get_base_url(), 'executor.queryProgress'),
            id=task_id,
            dict_data=dict_data,
            headers=req_header,
            stream=True,
            method='post',
        )

        image = None
        for stream_line in task_progress.iter_lines():
            stream_msg = json.loads(stream_line)
            t = stream_msg.get('type')

            code = stream_msg.get('code')
            if code and code >= 400:
                error = stream_msg.get('message')
                raise Exception(f'{ error or "Unknown Error"} session_id: {session_id}')

            if t == 'error':
                error = stream_msg.get('message')
                raise Exception(f'{ error or "Unknown Error"} session_id: {session_id}')

            elif t == 'report':
                status = stream_msg.get('status')
                if status == 'pending':
                    if replay:
                        console.log(
                            f'Cloud pending ... [dim]: {t} {task_id} ({status})[/dim]'
                        )
                    else:
                        st.update(f'Cloud pending ... [dim]: {t} ({status})[/dim]')

                elif status == 'failed':
                    error = stream_msg.get('error', {})
                    msg = error.get('message')
                    message = stream_msg.get('message')
                    raise Exception(
                        f'{ msg or message or "Unknown Error"} session_id: {session_id}'
                    )

                elif status == 'waiting':
                    error = stream_msg.get('error')
                    if replay:
                        task = stream_msg.get('task', {})
                        task_id = task.get('_id')
                        console.log(
                            f'Cloud waiting ... [dim]: {t} task: {task_id} {error or ""} ({status})[/dim]'
                        )
                    else:
                        st.update(
                            f'Cloud waiting ... [dim]: {t} {error or ""} ({status})[/dim]'
                        )

                elif status == 'succeeded':
                    if stream_msg.get('result', None):
                        image = stream_msg['result']
                        executor_id = image.get('id')
                        if replay:
                            console.log(
                                f'Cloud succeeded ... [dim]: {t} Executor: {executor_id} ({status})[/dim]'
                            )
                        else:
                            st.update(
                                f'Cloud succeeded ... [dim]: {t} Executor: {executor_id} ({status})[/dim]'
                            )
                    else:
                        task = stream_msg.get('task', {})
                        task_id = task.get('_id')
                        if replay:
                            console.log(
                                f'Cloud succeeded ... [dim]: {t} task: {task_id} ({status})[/dim]'
                            )
                        else:
                            st.update(
                                f'Cloud succeeded ... [dim]: {t} task: {task_id} ({status})[/dim]'
                            )

            elif t == 'progress':
                data = stream_msg.get('data', {})
                legacy_message = data.get('data', {})

                type = legacy_message.get('type')
                subject = legacy_message.get('subject')
                payload = legacy_message.get('payload', '')

                if verbose:
                    if type == 'console':
                        console.log(
                            f'Cloud building ... [dim]{subject}: {t} {payload}[/dim]'
                        )
                    else:
                        console.log(f'Cloud building ... [dim]{subject}: {type} [/dim]')
                else:
                    if type == 'console':
                        st.update(
                            f'Cloud building ... [dim]{subject}: {t} {payload}[/dim]'
                        )
                    else:
                        st.update(f'Cloud building ... [dim]{subject}: {type} [/dim]')

            else:
                console.log(f'Cloud succeeded ... [dim]: {t} [/dim]')

        return image

    def status(self) -> None:
        """Query the build status of the Executor."""

        task_id = None
        if self.args.id:
            task_id = self.args.id
        else:
            work_path = Path(self.args.path)
            uuid8, secret, exists_task_id = load_secret(work_path)
            task_id = exists_task_id

        if not task_id:
            raise Exception(
                'Error: Can\'t get task_id! You can set `--id your task_id` to get build progress info.'
            )

        verbose = True if self.args.verbose else False
        replay = True if self.args.replay else False

        console = get_rich_console()
        with console.status(f'Querying `{task_id}` ...') as st:

            image = self._status_with_progress(console, st, task_id, replay, verbose)

            if image:
                self.args.no_usage = False
                self._prettyprint_result(console, image)
            else:
                console.log(f'Waiting `{task_id}` ...')

    def _prettyprint_list_usage(self, console, executors, base_path):
        from rich import box
        from rich.panel import Panel
        from rich.table import Column, Table

        param_str = Table(
            Column(header="Executor", no_wrap=True),
            Column(header="Tag", no_wrap=True),
            Column(header="Relative path", no_wrap=True),
            box=box.SIMPLE,
            title=f'Base path: {base_path}',
            title_style='blue',
            show_lines=True,
            min_width=30 + len(str(base_path)),
        )

        for item in executors:
            name = item['name']
            relative_path = item['relative_path']
            tag = item.get('tag', None)
            param_str.add_row(
                f'{name}',
                f'{tag}' if tag else '',
                f'{relative_path}',
            )

        console.print(
            Panel(
                param_str,
                title='List',
                expand=False,
                width=120,
            )
        )

    def list(self) -> None:
        executors = []
        base_path = get_hub_packages_dir()
        for executor_dist_info_path in list_local():
            executor_path = executor_dist_info_path.parent
            config = load_config(executor_path)
            tag = get_tag_from_dist_info_path(executor_dist_info_path)
            relative_executor_path = executor_path.stem
            executors.append(
                {
                    'name': config['jtype'],
                    'tag': tag,
                    'relative_path': f'./{relative_executor_path}',
                }
            )

        console = get_rich_console()
        self._prettyprint_list_usage(console, executors, base_path)

    @staticmethod
    @disk_cache_offline(cache_file=str(get_cache_db()))
    def fetch_meta(
        name: str,
        tag: str,
        image_required: bool = True,
        rebuild_image: bool = True,
        *,
        secret: Optional[str] = None,
        force: bool = False,
    ) -> HubExecutor:
        """Fetch Executor metadata from Jina Hub.
        :param name: the UUID/name of the Executor
        :param tag: the tag of the Executor if available, otherwise, use `None` as the value
        :param secret: the access secret of the Executor
        :param image_required: indicates whether a Docker image is required or not
        :param rebuild_image: indicates whether Jina Hub needs to rebuild image or not
        :param force: if set to True, access to fetch_meta will always pull latest Executor metas, otherwise, default
            to local cache
        :return: meta of Executor

        .. note::
            Significant parameters like `name` and `tag` should be passed via ``args``
            and `force` and `secret` as ``kwargs``, otherwise, caching will not work.
        """
        import requests

        @retry(num_retry=3)
        def _send_request_with_retry(url, **kwargs):
            resp = requests.post(url, **kwargs)
            if resp.status_code != 200:
                if resp.text:
                    raise Exception(resp.text)
                resp.raise_for_status()

            return resp

        pull_url = urljoin(hubble.utils.get_base_url(), 'executor.getPackage')

        payload = {'id': name, 'include': ['code'], 'rebuildImage': rebuild_image}
        if image_required:
            payload['include'].append('docker')
        if secret:
            payload['secret'] = secret
        if tag:
            payload['tag'] = tag

        req_header = get_request_header()

        resp = _send_request_with_retry(pull_url, json=payload, headers=req_header)
        resp = get_json_from_response(resp)['data']

        images = resp['package'].get('containers', [])
        image_name = images[0] if images else None
        if image_required and not image_name:
            raise RuntimeError(
                f'No image found for Executor "{name}", '
                f'tag: {tag}, commit: {resp.get("commit", {}).get("id")}, '
                f'session_id: {req_header.get("jinameta-session-id")}'
            )
        buildEnv = resp['commit'].get('commitParams', {}).get('buildEnv', None)
        return HubExecutor(
            uuid=resp['id'],
            name=resp.get('name', None),
            commit_id=resp['commit'].get('id'),
            tag=tag or resp['commit'].get('tags', [None])[0],
            visibility=resp['visibility'],
            image_name=image_name,
            archive_url=resp['package']['download'],
            md5sum=resp['package']['md5'],
            build_env=list(buildEnv.keys()) if buildEnv else [],
        )

    @staticmethod
    def deploy_public_sandbox(args: Union[argparse.Namespace, Dict]) -> str:
        """
        Deploy a public sandbox to Jina Hub.
        :param args: arguments parsed from the CLI

        :return: the host and port of the sandbox
        """

        try:
            from jina import __version__ as jina_version
        except ImportError:
            jina_version = __unset_msg__

        args_copy = copy.deepcopy(args)
        if not isinstance(args_copy, Dict):
            args_copy = vars(args_copy)

        scheme, name, tag, secret = parse_hub_uri(args_copy.pop('uses', ''))
        payload = {
            'name': name,
            'tag': tag if tag else 'latest',
            'jina': jina_version,
            'args': args_copy,
            'secret': secret,
        }

        import requests

        console = get_rich_console()

        host = None
        port = None

        headers = get_request_header()
        response = requests.post(
            url=urljoin(hubble.utils.get_base_url(), 'sandbox.get'),
            json=payload,
            headers=headers,
        )
        json_response = get_json_from_response(response)
        if json_response.get('code') == 200:
            host = json_response.get('data', {}).get('host', None)
            port = json_response.get('data', {}).get('port', None)

        if host and port:
            console.log('🎉 A sandbox already exists, reusing it.')
            return host, port

        with console.status(
            f'[bold green]🚧 Deploying sandbox for [bold white]{name}[/bold white] since none exists...'
        ):
            try:
                response = requests.post(
                    url=urljoin(hubble.utils.get_base_url(), 'sandbox.create'),
                    json=payload,
                    headers=headers,
                )
                json_response = get_json_from_response(response)

                data = json_response.get('data') or {}
                host = data.get('host', None)
                port = data.get('port', None)
                if not host or not port:
                    raise Exception(f'Failed to deploy sandbox: {json_response}')

                console.log('🎉 Deployment completed, using it.')
            except BaseException:
                console.log(
                    'issue: https://github.com/floralatin/floralatin-hubble-sdk5/issues/new'
                )
                raise

        return host, port

    def _pull_with_progress(self, log_streams, console):
        from rich.progress import BarColumn, DownloadColumn, Progress

        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            DownloadColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            console=console,
            transient=True,
        ) as progress:
            tasks = {}
            for log in log_streams:
                if 'status' not in log:
                    continue
                status = log['status']
                status_id = log.get('id', None)
                pg_detail = log.get('progressDetail', None)

                if (pg_detail is None) or (status_id is None):
                    self.logger.debug(status)
                    continue

                if status_id not in tasks:
                    tasks[status_id] = progress.add_task(status, total=0)

                task_id = tasks[status_id]

                if ('current' in pg_detail) and ('total' in pg_detail):
                    progress.update(
                        task_id,
                        completed=pg_detail['current'],
                        total=pg_detail['total'],
                        description=status,
                    )
                elif not pg_detail:
                    progress.update(task_id, advance=0, description=status)

    def _load_docker_client(self):
        # with ImportExtensions(required=True):
        import docker.errors
        from docker import APIClient
        from hubble import __windows__

        try:
            self._client = docker.from_env()
            # low-level client
            self._raw_client = APIClient(
                base_url=docker.constants.DEFAULT_NPIPE
                if __windows__
                else docker.constants.DEFAULT_UNIX_SOCKET
            )
        except docker.errors.DockerException:
            self.logger.critical(
                'Docker daemon doesn\'t seem to be running. Please run the Docker daemon and try again.'
            )
            exit(1)

    def pull(self) -> str:
        """Pull the Executor package from Jina Hub.

        :return: the `uses` string
        """

        console = get_rich_console()
        cached_zip_file = None
        executor_name = None
        build_env = None

        try:
            need_pull = self.args.force_update
            with console.status(f'Pulling {self.args.uri}...') as st:
                scheme, name, tag, secret = parse_hub_uri(self.args.uri)
                image_required = scheme == 'jinahub+docker'

                st.update(f'Fetching [bold]{name}[/bold] from Jina Hub ...')
                executor, from_cache = HubIO.fetch_meta(
                    name,
                    tag,
                    image_required,
                    secret=secret,
                    force=need_pull,
                )

                build_env = executor.build_env

                presented_id = executor.name if executor.name else executor.uuid
                executor_name = (
                    f'{presented_id}'
                    if executor.visibility == 'public' or not secret
                    else f'{presented_id}:{secret}'
                ) + (f'/{tag}' if tag else '')

                if scheme == 'jinahub+docker':
                    self._load_docker_client()
                    import docker

                    try:
                        self._client.images.get(executor.image_name)
                    except docker.errors.ImageNotFound:
                        need_pull = True

                    if need_pull:
                        st.update('Pulling image ...')
                        log_stream = self._raw_client.pull(
                            executor.image_name, stream=True, decode=True
                        )
                        st.stop()
                        self._pull_with_progress(
                            log_stream,
                            console,
                        )
                    return f'docker://{executor.image_name}'
                elif scheme == 'jinahub':
                    import filelock

                    if build_env:
                        self._prettyprint_build_env_usage(console, build_env)

                    with filelock.FileLock(get_lockfile(), timeout=-1):
                        try:
                            pkg_path, pkg_dist_path = get_dist_path_of_executor(
                                executor
                            )
                            # check commit id to upgrade
                            commit_file_path = (
                                pkg_dist_path / f'PKG-COMMIT-{executor.commit_id or 0}'
                            )
                            if (not commit_file_path.exists()) and any(
                                pkg_dist_path.glob('PKG-COMMIT-*')
                            ):
                                raise FileNotFoundError(
                                    f'{pkg_path} need to be upgraded'
                                )

                            st.update(
                                'Installing dependencies from [bold]requirements.txt[/bold]...'
                            )
                            install_package_dependencies(
                                install_deps=self.args.install_requirements,
                                pkg_dist_path=pkg_dist_path,
                                pkg_path=pkg_dist_path,
                            )

                        except FileNotFoundError:
                            need_pull = True

                        if need_pull:
                            # pull the latest Executor meta, as the cached data would expire
                            if from_cache:
                                executor, _ = HubIO.fetch_meta(
                                    name,
                                    tag,
                                    image_required,
                                    secret=secret,
                                    force=True,
                                )

                            st.update(f'Downloading {name} ...')
                            cached_zip_file = download_with_resume(
                                executor.archive_url,
                                get_download_cache_dir(),
                                f'{executor.uuid}-{executor.md5sum}.zip',
                                md5sum=executor.md5sum,
                            )

                            st.update(f'Unpacking {name} ...')
                            install_local(
                                cached_zip_file,
                                executor,
                                install_deps=self.args.install_requirements,
                            )

                            pkg_path, _ = get_dist_path_of_executor(executor)

                        return f'{pkg_path / "config.yml"}'
                else:
                    raise ValueError(f'{self.args.uri} is not a valid scheme')
        except KeyboardInterrupt:
            executor_name = None
        except Exception:
            executor_name = None
            raise
        finally:
            # delete downloaded zip package if existed
            if cached_zip_file is not None:
                cached_zip_file.unlink()

            if not self.args.no_usage and executor_name:
                self._prettyprint_usage(console, executor_name)



import os

from . import Client
from . import login as _login
from . import logout as _logout
from .excepts import AuthenticationFailedError, AuthenticationRequiredError
from .utils.api_utils import get_json_from_response


def logout(*args):
    _logout()


def login(args):
    client = Client(jsonify=True)
    nickname = ''

    if args.force:
        _login(prompt='login', force=True)
        return

    try:
        nickname = client.username
    except Exception as ex:
        if isinstance(ex, (AuthenticationRequiredError, AuthenticationFailedError)):
            _login(prompt='login')
        else:
            raise ex
    else:
        from .dockerauth import auto_deploy_hubble_docker_credential_helper

        auto_deploy_hubble_docker_credential_helper()

        if nickname:
            from rich.console import Console

            console = Console()
            console.print(
                f':closed_lock_with_key: You are already logged in as [b green]{nickname}[/b green].',
                '',
                'If you want to log in to another account, please run either:',
                '- [b]jina auth logout[/]',
                '- or, [b]jina auth login -f[/]',
                sep=os.linesep,
            )


def token(args):
    client = Client(max_retries=None)

    if args.operation == 'create':
        response = client.create_personal_access_token(
            name=args.name, expiration_days=args.expire
        )

        response.raise_for_status()
        json_response = get_json_from_response(response)
        token = json_response['data']['token']

        import rich
        from rich.panel import Panel

        rich.print(
            Panel(
                f'''[b]{token}[/b]

You can set it as an env var [b]JINA_AUTH_TOKEN[/b]''',
                title=':party_popper: [green]New token created[/]',
                subtitle=':point_up:️ [yellow] This token is only shown once![/]',
                width=50,
            )
        )

    if args.operation == 'delete':
        response = client.delete_personal_access_token(name=args.name)

        response.raise_for_status()

        import rich
        from rich.panel import Panel

        rich.print(
            Panel(
                f'[b]{args.name}[/b]',
                title=':party_popper: [green]Successfully deleted',
                width=50,
            )
        )

    if args.operation == 'list':
        response = client.list_personal_access_tokens()
        response.raise_for_status()
        json_response = get_json_from_response(response)
        tokens = json_response['data']['personal_access_tokens']

        import rich
        from rich.table import Table

        table = Table(title='Your Personal Access Tokens', highlight=True)
        table.add_column('Name')
        table.add_column('Type')
        table.add_column('Create at', justify='center')
        table.add_column('Expire at', justify='center')
        table.add_column('Last use at', justify='center')

        for token in tokens:
            table.add_row(
                token['name'],
                token['type'],
                token['createdAt'],
                token['expireAt'],
                token['updatedAt'],
            )

        rich.print(table)



from typing import IO, Any, MutableMapping, Optional, Text, Union

import requests

from ..excepts import errorcodes
from ..utils.api_utils import get_base_url, get_json_from_response
from ..utils.auth import Auth
from .session import HubbleAPISession


class BaseClient(object):
    """Base Hubble Python API client.

    :param max_retries: Number of allowed maximum retries.
    :param jsonify: Convert `requests.Response` object to json.
    """

    def __init__(
        self,
        token: Optional[str] = None,
        max_retries: Optional[int] = None,
        jsonify: bool = False,
    ):
        self._session = HubbleAPISession()

        self._token = token if token else Auth.get_auth_token()
        if self._token:
            self._session.init_jwt_auth(token=self._token)

        self._base_url = get_base_url()
        self._jsonify = jsonify
        if max_retries:
            from requests.adapters import HTTPAdapter

            self._session.mount(
                prefix=self._base_url, adapter=HTTPAdapter(max_retries=max_retries)
            )

    def _handle_error_request(self, resp: Union[requests.Response, dict]):
        if isinstance(resp, requests.Response):
            resp = get_json_from_response(resp)

        message = resp.get('message', None)
        code = resp.get('status', -1)
        data = resp.get('data', {})

        ExceptionCls = errorcodes[code]

        raise ExceptionCls(response=resp, data=data, message=message, code=code)

    def handle_request(
        self,
        url: str,
        method: str = 'POST',
        data: Optional[dict] = None,
        files: Optional[MutableMapping[Text, IO[Any]]] = None,
        headers: Optional[dict] = None,
        json: Optional[dict] = None,
    ) -> Union[requests.Response, dict]:
        """The basis request handler.

        Hubble API consider all requests as POST requests.
        The method leverages the ``HubbleAPISession`` to send
        POST requests based on parameters.

        :param url: The url of the request.
        :param method: The request type, for v2 always set to POST.
        :param data: Optional data payloads to be send along with request.
        :param files: Optional files to be uploaded.
        :returns: `requests.Response` object as returned value
            or dict if jsonify.
        """
        resp = self._session.request(
            method=method,
            url=url,
            data=data if data else None,
            files=files,
            headers=headers,
            json=json if json else None,
        )
        if resp.status_code >= 400:
            self._handle_error_request(resp)

        if self._jsonify:
            resp = get_json_from_response(resp)

        return resp



"""Argparser module for hub list"""

from hubble.parsers.helper import add_arg_group


def mixin_hub_list_parser(parser):
    """Add the arguments for hub list to the parser
    :param parser: the parser configure
    """

    add_arg_group(parser, title='List')



from setuptools import find_packages, setup

# package metadata
_description = 'SDK for Hubble API at Jina AI.'
_setup_requires = ['setuptools>=18.0', 'wheel']
_python_requires = '>=3.7.0'
_author = 'Jina AI'
_email = 'hello@jina.ai'
_keywords = (
    'jina neural-search neural-network deep-learning pretraining '
    'fine-tuning pretrained-models triplet-loss metric-learning '
    'siamese-network few-shot-learning'
)
_url = 'https://github.com/jina-ai/hubble-client-python/'
_download_url = 'https://github.com/jina-ai/hubble-client-python/tags'
_classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Environment :: Console',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
]
_project_urls = {
    'Source': 'https://github.com/jina-ai/hubble-client-python/',
    'Tracker': 'https://github.com/jina-ai/hubble-client-python/issues',
}
_license = 'Proprietary'
_package_exclude = ['*.tests', '*.tests.*', 'tests.*', 'tests']


# package requirements
try:
    with open('requirements.txt', 'r') as f:
        _main_deps = f.readlines()
except FileNotFoundError:
    _main_deps = []


try:
    with open('requirements-dev.txt', 'r') as f:
        _extra_deps = {'full': f.read().splitlines()}
        _extra_deps['full'].remove('-r requirements.txt')
        _extra_deps['full'].extend(_main_deps)
except FileNotFoundError:
    _extra_deps = {}

# package long description
try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''


if __name__ == '__main__':
    setup(
        name='floralatin-hubble-sdk5',
        packages=find_packages(exclude=_package_exclude),
        include_package_data=True,
        description=_description,
        author=_author,
        author_email=_email,
        url=_url,
        license=_license,
        download_url=_download_url,
        long_description=_long_description,
        long_description_content_type='text/markdown',
        zip_safe=False,
        setup_requires=_setup_requires,
        install_requires=_main_deps,
        extras_require=_extra_deps,
        python_requires=_python_requires,
        classifiers=_classifiers,
        project_urls=_project_urls,
        keywords=_keywords,
        entry_points={
            'console_scripts': [
                'jina-auth=hubble.__main__:main',
                'jina-hub=hubble.executor.__main__:main',
                'docker-credential-jina-hubble=hubble.dockerauth:main',
            ],
        },
    )



import json
import os
import webbrowser
from typing import Callable, Optional
from urllib.parse import urlencode, urljoin

import aiohttp
import requests
from hubble.client.session import HubbleAPISession
from hubble.excepts import AuthenticationFailedError
from hubble.utils.api_utils import get_base_url, get_json_from_response
from hubble.utils.config import config
from rich import print as rich_print

JINA_LOGO = (
    'https://d2vchdhjlcm3i6.cloudfront.net/Company+Logo/Light/Company+logo_light.svg'
)

NOTEBOOK_LOGIN_HTML = f"""
<div class='custom-container'>
    <style>
        .button1 {{
            color: white;
            background-color: #009191;
            border: 1px solid #009191;
        }}
        .button2 {{
            color: #009191;
            background-color: white;
            border: 1px solid #009191;
        }}
        .link1 {{
            color:#009191;
            position: relative;
            top: 22px;
            right: -120px;
            z-index: 99;
        }}
        .custom-container {{
            margin-top: 10px;
            margin-bottom: -10px;
        }}
        .spaced {{
            margin: 20px 0;
        }}
    </style>
    <center>
        <img src={JINA_LOGO} width=175 alt='Jina AI'>
        <div class='spaced'></div>
        <p>
            Copy a <b>Personal Access Token</b>, paste it below, and press the <b>Token login</b> button.
            <br>
            If you don't have a token, press the <b>Browser login</b> button to log in via the browser.
        </p>
        <a
            href='https://hub.jina.ai/user/tokens'
            target='__blank'
            class='link1'>
                Create
        </a>
    </center>
</div>
"""

NOTEBOOK_SUCCESS_HTML = f"""
<div class='custom-container'>
    <style>
        .custom-container {{
            margin-top: 10px;
            margin-bottom: 0;
        }}
        .spaced {{
            margin: 20px 0;
        }}
    </style>
    <center>
        <img src={JINA_LOGO} width=175 alt='Jina AI'>
        <div class='spaced'></div>
        <p>
            You are logged in to Jina AI!
        </p>
        <p>
            If you want to log in again, run <code>notebook_login(force=True)</code>.
        </p>
    </center>
</div>
"""

NOTEBOOK_ERROR_HTML = """
<div class='custom-container'>
    <style>
        .custom-container {{
            margin-top: 10px;
            margin-bottom: 0;
        }}
        .spaced {{
            margin: 20px 0;
        }}
        .error {{
            text-align: left !important;
            background-color: WhiteSmoke;
            margin: 10px 0 !important;
            padding: 10px 50px 10px 20px;
            line-height: 16px;
        }}
        .red {{
            color: #d03c38;
        }}
    </style>
    <center>
        <img src={LOGO} width=175 alt='Jina AI'>
        <div class='spaced'></div>
        <p class='red'>
            An error occured, see the details below.
        </p>
        <div class='error'>
            <pre><code>{ERR}</code></pre>
        </div>
    </center>
</div>
"""

NOTEBOOK_REDIRECT_HTML = """
<div class='custom-container'>
    <style>
        .custom-container {{
            margin-top: 10px;
            margin-bottom: 0;
        }}
        .spaced {{
            margin: 20px 0;
        }}
    </style>
    <center>
        <img src={LOGO} width=175 alt="Jina AI">
        <div class='spaced'></div>
        <p>
            Please open <a href='{HREF}' target='_blank'>this link</a> to continue the login process.
        </p>
    </center>
</div>
"""


class Auth:
    @staticmethod
    def get_auth_token_from_config():
        """Get user auth token from config file."""
        token_from_config: Optional[str] = None
        if isinstance(config.get('auth_token'), str):
            token_from_config = config.get('auth_token')

        return token_from_config

    @staticmethod
    def get_auth_token():
        """Get user auth token.

        .. note:: We first check `JINA_AUTH_TOKEN` environment variable.
          if token is not None, use env token. Otherwise, we get token from config.
        """
        token_from_env = os.environ.get('JINA_AUTH_TOKEN')

        token_from_config: Optional[str] = Auth.get_auth_token_from_config()

        return token_from_env if token_from_env else token_from_config

    @staticmethod
    def validate_token(token):
        try:
            session = HubbleAPISession()
            session.init_jwt_auth(token)
            resp = session.validate_token()
            resp.raise_for_status()
        except requests.exceptions.HTTPError:
            raise AuthenticationFailedError("Could not validate token")

    @staticmethod
    def login_notebook(
        force: bool = False, post_success: Optional[Callable] = None, **kwargs
    ):
        """Login user in notebook environments like colab"""

        # trying to import utilities (only available in notebook env)
        try:
            import ipywidgets.widgets as widgets
            from IPython.display import clear_output, display
        except ImportError:
            raise ImportError(
                """
The `notebook_login` function can only be used in a notebook.
The function also requires `ipywidgets`.
                """
            )

        # creating widgets
        # reusable layout, for all widgets
        layout = widgets.Layout(
            display="flex", flex_flow="column", align_items="center"
        )

        # login widget
        token_widget = widgets.Password(
            placeholder="Personal Access Token (PAT)",
            layer=widgets.Layout(width="300px"),
        )

        token_button_widget = widgets.Button(
            description="Token login",
            disabled=True,
            layout=widgets.Layout(width="300px"),
        )

        token_button_widget.add_class('button1')

        def _handle_token_change(change):
            if change.new is not None and change.new != '':
                token_button_widget.disabled = False
            else:
                token_button_widget.disabled = True

        token_widget.observe(_handle_token_change, names='value')

        browser_button_widget = widgets.Button(
            description="Browser login",
            layout=widgets.Layout(width="300px", margin="10px 0 0 0"),
        )

        browser_button_widget.add_class('button2')

        login_widget = widgets.VBox(
            [
                widgets.HTML(NOTEBOOK_LOGIN_HTML),
                token_widget,
                token_button_widget,
                browser_button_widget,
            ],
            layout=layout,
        )

        # sucess widget
        success_widget = widgets.VBox(
            [widgets.HTML(NOTEBOOK_SUCCESS_HTML)], layout=layout
        )

        # redirect url widget
        redirect_url_widget = widgets.HTML(value="")
        redirect_widget = widgets.VBox(
            [
                redirect_url_widget,
            ],
            layout=layout,
        )

        # error widget
        error_description_widget = widgets.HTML(value="")
        error_widget = widgets.VBox(
            [
                error_description_widget,
            ],
            layout=layout,
        )

        # callback functions for login_async to communicate events
        def _success_callback(**kwargs):
            clear_output()
            display(success_widget)
            if post_success:
                post_success()

        def _redirect_callback(href=None, **kwargs):
            redirect_url_widget.value = NOTEBOOK_REDIRECT_HTML.format(
                LOGO=JINA_LOGO, HREF=href
            )
            clear_output()
            display(redirect_widget)

        def _error_callback(err=None, **kwargs):
            error_description_widget.value = NOTEBOOK_ERROR_HTML.format(
                LOGO=JINA_LOGO, ERR=err
            )
            clear_output()
            display(error_widget)

        # login function called when pressing the login button
        def _login(*args):

            # reading token, clearing form, disabling elements
            token = token_widget.value
            token_widget.value = ""
            token_widget.disabled = True
            token_button_widget.disabled = True
            browser_button_widget.disabled = True

            # verify token before login function
            if token != "":
                try:
                    Auth.validate_token(token)
                    config.set('auth_token', token)
                    clear_output()
                    display(success_widget)
                    return
                except AuthenticationFailedError:
                    pass

            Auth.login_sync(
                force=force,
                success_callback=_success_callback,
                redirect_callback=_redirect_callback,
                error_callback=_error_callback,
                **kwargs,
            )

        token_button_widget.on_click(_login)
        browser_button_widget.on_click(_login)

        # verifying existing token
        token = Auth.get_auth_token()
        if token and not force:
            try:
                Auth.validate_token(token)
                clear_output()
                display(success_widget)
                return
            except AuthenticationFailedError:
                pass

        # show login widget
        display(login_widget)

    @staticmethod
    def login_sync(
        force=False,
        success_callback=None,
        redirect_callback=None,
        error_callback=None,
        **kwargs,
    ):
        # verify if token already exists, authenticate token if exists
        if not force:
            token = Auth.get_auth_token()
            if token:
                try:
                    Auth.validate_token(token)
                    if success_callback:
                        success_callback()
                    return
                except AuthenticationFailedError:
                    pass

        api_host = get_base_url()
        auth_info = None

        # authorize user
        url = urljoin(
            api_host,
            'user.identity.proxiedAuthorize?{}'.format(
                urlencode({'provider': 'jina-login'})
            ),
        )

        response = requests.get(url, stream=True)

        # iterate through response
        for line in response.iter_lines():
            item = json.loads(line.decode('utf-8'))
            event = item['event']

            if event == 'redirect':
                href = item['data']["redirectTo"]
                if redirect_callback:
                    redirect_callback(href=href)
                else:
                    print(f'Please open the following link: {href}')

            elif event == 'authorize':
                if item['data']['code'] and item['data']['state']:
                    auth_info = item['data']
                else:
                    err = item['data']["error_description"]
                    if error_callback:
                        error_callback(err=err)
                    else:
                        print('Authentication failed: {}'.format(err))

            elif event == 'error':
                err = item['data']
                if error_callback:
                    err = json.dumps(err, indent=4)
                    error_callback(err=err)
                else:
                    print('Authentication failed: {}'.format(err))
            else:
                err = f'Unknown event: {event}'
                if error_callback:
                    error_callback(err=err)
                else:
                    print(err)

        if auth_info is None:
            return

        # retrieving and saving token
        url = urljoin(api_host, 'user.identity.grant.auto')
        response = requests.post(url, json=auth_info)
        response.raise_for_status()
        json_response = get_json_from_response(response)
        token = json_response['data']['token']
        config.set('auth_token', token)

        # dockerauth
        from hubble.dockerauth import auto_deploy_hubble_docker_credential_helper

        auto_deploy_hubble_docker_credential_helper()

        if success_callback:
            success_callback()

    @staticmethod
    async def login(force=False, **kwargs):
        # verify if token already exists, authenticate token if exists
        if not force:
            token = Auth.get_auth_token()
            if token:
                try:
                    Auth.validate_token(token)
                    return
                except AuthenticationFailedError:
                    pass

        api_host = get_base_url()
        auth_info = None
        async with aiohttp.ClientSession(trust_env=True) as session:
            kwargs['provider'] = kwargs.get('provider', 'jina-login')

            async with session.get(
                url=urljoin(
                    api_host,
                    'user.identity.proxiedAuthorize?{}'.format(urlencode(kwargs)),
                ),
            ) as response:
                async for line in response.content:
                    item = json.loads(line.decode('utf-8'))
                    event = item['event']
                    if event == 'redirect':
                        print(
                            f'Your browser is going to open the login page.\n'
                            f'If this fails please open the following link: {item["data"]["redirectTo"]}'
                        )
                        webbrowser.open(item['data']['redirectTo'])
                    elif event == 'authorize':
                        if item['data']['code'] and item['data']['state']:
                            auth_info = item['data']
                        else:
                            rich_print(
                                ':rotating_light: Authentication failed: {}'.format(
                                    item['data']['error_description']
                                )
                            )
                    elif event == 'error':
                        rich_print(
                            ':rotating_light: Authentication failed: {}'.format(
                                item['data']
                            )
                        )
                    else:
                        rich_print(':rotating_light: Unknown event: {}'.format(event))

        if auth_info is None:
            return

        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.post(
                url=urljoin(api_host, 'user.identity.grant.auto'),
                data=auth_info,
            ) as response:
                response.raise_for_status()
                json_response = await response.json()
                token = json_response['data']['token']
                user = json_response['data']['user']['nickname']

                config.set('auth_token', token)

                from hubble.dockerauth import (
                    auto_deploy_hubble_docker_credential_helper,
                )

                auto_deploy_hubble_docker_credential_helper()

                rich_print(
                    f':closed_lock_with_key: [green]Successfully logged in to Jina AI[/] as [b]{user}[/b]!'
                )

    @staticmethod
    async def logout():
        api_host = get_base_url()

        token = Auth.get_auth_token()
        token_from_config = Auth.get_auth_token_from_config()
        if token != token_from_config:
            rich_print(':warning: The token from environment variable is ignored.')

        async with aiohttp.ClientSession(trust_env=True) as session:
            session.headers.update({'Authorization': f'token {token_from_config}'})

            async with session.post(
                url=urljoin(api_host, 'user.session.dismiss')
            ) as response:
                json_response = await response.json()
                if json_response['code'] == 401:
                    rich_print(
                        ':unlock: You are not logged in locally. There is no need to log out.'
                    )
                elif json_response['code'] == 200:
                    rich_print(':unlock: You have successfully logged out.')
                    config.delete('auth_token')
                else:
                    rich_print(
                        f':rotating_light: Failed to log out. {json_response["message"]}'
                    )



from dataclasses import dataclass


@dataclass(frozen=True)
class EndpointsV2(object):
    """All available Hubble API endpoints."""

    initiate_session_authorize: str = 'user.identity.authorize'
    initiate_proxied_authorize: str = 'user.identity.proxiedAuthorize'
    auto_grant_user_identity: str = 'user.identity.grant.auto'
    get_user_info: str = 'user.identity.whoami'
    dismiss_user_session: str = 'user.session.dismiss'

    create_pat: str = 'user.pat.create'
    list_pats: str = 'user.pat.list'
    delete_pat: str = 'user.pat.delete'

    upload_artifact: str = 'artifact.upload'
    download_artifact: str = 'artifact.getDownloadUrl'
    delete_artifact: str = 'artifact.delete'
    get_artifact_info: str = 'artifact.getDetail'
    list_artifacts: str = 'artifact.list'

    list_internal_docker_registries: str = 'dockerRegistry.listInternalRegistries'
